# Copyright (C) 2003-2011 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software 
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

"""Provide an authentication wrapper around htdig's htsearch.

In order to search a list's archive the HTTP request must present a valid 
cookie authorizing access to the list's archives for private archives. 
This cookie must be obtained by the same process as the user must 
adopt for accessing the archive via the private.py script.

The basic premise is that htsearch cannot reach a list's htdig conf file
unless this script enhances the environment in which htsearch runs to
provide access to a list's htdig conf file.

If the use authenticates OK for a private archive, or the archive is public,
this script:

a. in the case of local htdig, this script enhances the environment with 
'CONFIG-DIR' pointing to MM's list htdig conf files, and runs htsearch.

b. in the case of remote htdig, this script performs a HTTP request to a
companion CGI script running on the htdig machine. This request can of
course be forged easily. The only protection available is the the companion
script can optionaaly check the IP number from which requests to it are made
and reject them if they do not come from the MM server.
"""

import sys
import os
import cgi
import re
import exceptions
import types
import httplib
import urllib
import urlparse

from subprocess import Popen, PIPE

from Mailman import mm_cfg
from Mailman import Utils
from Mailman import MailList
from Mailman import Errors
from Mailman import i18n
from Mailman.htmlformat import *
from Mailman.Logging.Syslog import syslog


# Set up i18n.  Until we know which list is being requested, we use the
# server's default.
_ = i18n._
i18n.set_language(mm_cfg.DEFAULT_SERVER_LANGUAGE)

errors = {'cgi': _('CGI problem.'),
          'info': _('Path info.'),
          'list': _('The requested list cannot be accessed.'),
          'htsearch': _('search failed'),
          'auth': Utils.maketext('htdig_auth_failure.html', dict=None, raw=0)
         }

class _search_exception(exceptions.Exception):
    
    def __init__(self, listname, reason, detail):
        self.listname = Utils.websafe(listname)
        self.reason = reason 
        self.detail = Utils.websafe(detail)

    def __str__(self):
        return 'listname: %s, reason: %s, detail: %s' % (self.listname, self.reason, self.detail)


def true_path(path):
    "Ensure that the path is safe by removing .."
    path = path.replace("../", "")
    path = path.replace("./", "")
    return path[1:]

def make_inserts(listname):
    urlbase =  mm_cfg.DEFAULT_URL or \
               mm_cfg.DEFAULT_URL_PATTERN % mm_cfg.DEFAULT_URL_HOST
    return {
            'mailto': Utils.get_site_email(),
            'listinfo_link': urlbase + '/listinfo/' + listname,
            'referer': Utils.websafe(
                os.environ.get('HTTP_REFERER', _('Referer not known'))),
            'uri': Utils.websafe(
                os.environ.get('REQUEST_URI', _('URI not known'))),
           }

def error_quit(why):
    d = make_inserts(why.listname)
    d['error'] = errors[why.reason] + ' ' + why.detail
    charset = Utils.GetCharSet(mm_cfg.DEFAULT_SERVER_LANGUAGE)
    if why.reason == 'list':
        stuff = 'Status: 404 Not Found\n'
    elif why.reason == 'auth':
        stuff = 'Status: 401 Unauthorized\n'
    else:
        stuff = ''
    stuff += 'Content-type: text/html; charset=' + charset + '\n\n'
    stuff += Utils.maketext('htdig_access_error.html', dict=d, 
                         lang=mm_cfg.DEFAULT_SERVER_LANGUAGE)
    print stuff
    syslog('error', "htsearch for list: %s, cause: %s, detail: %s" % \
           (why.listname, why.reason, why.detail))
    sys.exit(0)

_required_fields = ('method',
                    'format',
                    'sort',
                    'config',
                   )

_allowed_fields = {'method': 0,
                   'format': 0,
                   'sort': 0,
                   'config': 1,
                   'words': 0,
                   'submit': 0,
                   'restrict': 0,
                   'exclude': 0,
                   'page': 0,
                  }

def check_params(listname, fs, detail):
    detail = detail + ' fields: ' + ','.join(fs.keys())
    for fieldname in _required_fields:
        if not fs.has_key(fieldname):
            raise _search_exception(listname, 'cgi',  '-5-' + detail)
    fieldhash = {}
    for fieldname in fs.keys():
        if not _allowed_fields.has_key(fieldname):
            raise _search_exception(listname, 'cgi', '-6- ' + detail)
        if  type(fs[fieldname]) is types.ListType:
            raise _search_exception(listname, 'cgi', '-8- ' + detail)        
        fieldhash[fieldname] = fs[fieldname].value
    return urllib.urlencode(fieldhash)

def main():
    try:
        try:
            request_method = os.environ['REQUEST_METHOD']
            form = cgi.FieldStorage()
        except:
            raise _search_exception('', 'cgi', 'No list -1-')
        list_info = Utils.GetPathPieces()
        if not list_info or len(list_info) != 1:
            raise _search_exception('', 'info', 'No list -2-')
        path_listname = list_info[0].lower()
        # Check right number of expected fields and get them
        # url encoded
        params = check_params(path_listname, form, 'Field count -4-')
        # Extract the listname from the form
        listname = form['config'].value
        if listname.endswith('.htsearch'):
            listname = listname[:-len('.htsearch')]
        # Access the list OK?
        if path_listname != listname:
            raise _search_exception(listname, 'list', '%s:%s' % (path_listname, listname))
        try:
            mlist = MailList.MailList(listname, lock=0)
        except:
            raise _search_exception(listname, 'list', '-11-')
        # We only need to authorize the user if it's a private archive 
        if mlist.archive_private:
            if not mlist.WebAuthenticate((mm_cfg.AuthUser,
                                             mm_cfg.AuthListModerator,
                                             mm_cfg.AuthListAdmin,
                                             mm_cfg.AuthSiteAdmin),
                                            '', ''):
                raise _search_exception(listname, 'auth', '-10-')
        if mm_cfg.REMOTE_HTDIG:
            #  We are going to do an HTTP request to the server that is
            # running htsearch and then return the response.
            headers = {"Content-type": "application/x-www-form-urlencoded",
                       "Accept": "text/plain, text/html"}
            urlbits = urlparse.urlsplit(mm_cfg.REMOTE_MMSEARCH_URL)
            hostbit = urlbits[1]
            uribit = urlbits[2] + '/' + listname
            conn = httplib.HTTPConnection(hostbit)
            conn.request("POST", uribit, params, headers)
            response = conn.getresponse()
            if response.status == 200:
                ctype = response.getheader('Content-type', 'text/html')
                print "Content-type: %s\n\n" % ctype
                print response.read()
                conn.close()
            else:
                syslog('error', "remote htsearch failed, url: %s, response code: %d" % \
                       (mm_cfg.REMOTE_MMSEARCH_URL, response.status))
                conn.close()
                raise _search_exception(listname, 'cgi', 
                                        ' -9- %d' % response.status)
        else:
            # We are going to execute htsearch and return its response.
            # We need to let htsearch get at the list specific htdig
            # conf file
            os.environ['CONFIG_DIR'] = mm_cfg.HTDIG_CONF_LINK_DIR
            cmd = "%s" % mm_cfg.HTDIG_HTSEARCH_PATH
            os.environ['CONTENT_LENGTH'] = str(len(params))
            child = Popen(cmd, stdin=PIPE, stdout=PIPE)
            if request_method == 'POST':
                response = child.communicate(params)[0]
            else:
                response = child.communicate()[0]
            exitstatus = child.returncode
            exitstatus = (exitstatus >> 8) & 0xff
            if exitstatus:
                syslog('error', "htsearch for list: %s, existatus: %s" % \
                       (listname, str(exitstatus)))
            if not response:
                raise _search_exception(listname, 'htsearch', '-12-') 
            print response
    except _search_exception, e:
        error_quit(e)
    sys.exit(0)
