# Copyright (C) 1998-2018 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""Extend mailbox.UnixMailbox.
"""

import sys
import mailbox

import email
from email.Parser import Parser
from email.Errors import MessageParseError

from Mailman import mm_cfg
from Mailman.Message import Generator
from Mailman.Message import Message

try:
    True, False
except NameError:
    True = 1
    False = 0



def _safeparser(fp):
    try:
        return email.message_from_file(fp, Message)
    except MessageParseError:
        # Don't return None since that will stop a mailbox iterator
        return ''



# Changed to stricter UnixMailbox from PortableUnixMailbox because
# unescaped From's in message bodies can be received by Mailman in the
# and this has caused problems with message extraction during archive 
# rebuilding when MHonnArc is used for archiving.
class Mailbox(mailbox.UnixMailbox):
    def __init__(self, fp):
        mailbox.UnixMailbox.__init__(self, fp, _safeparser)

    # msg should be an rfc822 message or a subclass.
    def AppendMessage(self, msg):
        # Check the last character of the file and write a newline if it isn't
        # a newline (but not at the beginning of an empty file).
        try:
            self.fp.seek(-1, 2)
        except IOError, e:
            # Assume the file is empty.  We can't portably test the error code
            # returned, since it differs per platform.
            pass
        else:
            if self.fp.read(1) <> '\n':
                self.fp.write('\n')
        # Seek to the last char of the mailbox
        self.fp.seek(0, 2)
        # Create a Generator instance to write the message to the file
        g = Generator(self.fp)
        g.flatten(msg, unixfrom=True)
        # Add one more trailing newline for separation with the next message
        # to be appended to the mbox.
        print >> self.fp



# This stuff is used by pipermail.py:processUnixMailbox().  It provides an
# opportunity for the built-in archiver to scrub archived messages of nasty
# things like attachments and such...
def _archfactory(mailbox):
    # The factory gets a file object, but it also needs to have a MailList
    # object, so the clearest <wink> way to do this is to build a factory
    # function that has a reference to the mailbox object, which in turn holds
    # a reference to the mailing list.  Nested scopes would help here, BTW,
    # but we can't rely on them being around (e.g. Python 2.0).
    def scrubber(fp, mailbox=mailbox):
        msg = _safeparser(fp)
        if msg == '':
            return msg
        return mailbox.scrub(msg)
    return scrubber


class ArchiverMailbox(Mailbox):
    # This is a derived class which is instantiated with a reference to the
    # MailList object.  It is build such that the factory calls back into its
    # scrub() method, giving the scrubber module a chance to do its thing
    # before the message is archived.
    def __init__(self, fp, mlist):
        if mm_cfg.ARCHIVE_SCRUBBER:
            __import__(mm_cfg.ARCHIVE_SCRUBBER)
            self._scrubber = sys.modules[mm_cfg.ARCHIVE_SCRUBBER].process
        else:
            self._scrubber = None
        self._mlist = mlist
        mailbox.UnixMailbox.__init__(self, fp, _archfactory(self))

    def scrub(self, msg):
        if self._scrubber:
            return self._scrubber(self._mlist, msg)
        else:
            return msg

    def skipping(self, flag):
        """ This method allows the archiver to skip over messages without
        scrubbing attachments into the attachments directory."""
        if flag:
            self.factory = _safeparser
        else:
            self.factory = _archfactory(self)

import tempfile
from Mailman.i18n import _
from Mailman import Errors
from Mailman.Logging.Syslog import syslog

class mhonarcArchiverMailbox(ArchiverMailbox):
    # This class is a hack to support integrated use of MHonArc for generating
    # the archive content. 
    # Primary purpose is to return either a single message object or a path
    # to a file containing a sequence of messages if thos messages should
    # all go into the same archive.
    # Value returned is a triple  of:
    #    type of info returned; 1: single message, 2: multiple messages
    #    a list of archives the message(s) should go in
    #    a single message object or the path to a temporary mbox file containing
    #    multiple messages

    def __init__(self, archiver, fp, mlist, start, end):
        ArchiverMailbox.__init__(self, fp, mlist)
        self._archiver = archiver
        self._input = fp
        self._mlist = mlist
        if start is None:
            start = 0
        self._start = start
        self._end = end
        self._heldover = []
        self._counter = 0
        self._finished = 0
        # Skip over initial, unwanted messages
        while self._counter < start:
            try:
                msg = ArchiverMailbox.next(self)
            except Errors.DiscardMessage:
                continue
            if msg is None:
                self._finished = 1
                break
            self._counter += 1
    
    # Leave scrubbing to MHonArc
    def scrub(self, msg):
        return msg
    
    def next(self):
        mbox_path = mbox_fp = None
        result = None
        try:
            while 1:
                if self._end is not None and self._counter >= self._end:
                    # Processed as many messages as required
                    self._finished = 1
                if self._finished:
                    # No more messages to be processed
                    if self._heldover:
                        # Want to return final tranche of messages
                        result = self._heldover
                        self._heldover = []
                    else:
                        # No more messages to return
                        result = None
                    break
                try:
                    pos = self._input.tell()
                    msg = ArchiverMailbox.next(self)
                except Errors.DiscardMessage:
                    continue
                except Exception:
                    syslog('error', 'uncaught archiver exception at filepos: %s',
                           pos)
                    if mbox_fp:
                        mbox_fp.close()
                        os.unlink(mbox_path)
                    self._heldover = []
                    self._finished = 1
                    raise
                self._counter += 1
                if msg is None:
                    # No messages left
                    self._finished = 1
                    continue
                if msg == '':
                    # It was an unparseable message
                    continue
                # We have a message 
                msgid = msg.get('message-id', 'n/a')
                counter = self._counter
                self._archiver.message(_('#%(counter)05d %(msgid)s'))
                article = self._archiver._makeMhonarcArticle(msg)
                # Determine which archive(s) the message should be archived to
                this_archive = self._archiver.get_archives(article)
                if type(this_archive) == type(''):
                    this_archive = [this_archive]
                # Now depends on if we are holding message(s) already
                if not self._heldover:
                    # Not holding message(s) so hold this one
                    self._heldover = [1, this_archive, msg]
                    continue
                else:
                    # We are holding a message
                    if this_archive != self._heldover[1]:
                        # New messge is not for the same archive(s)
                        result = self._heldover
                        # Hold over the new message
                        self._heldover = [1, this_archive, msg]
                        # Exit to the return what was held over
                        break
                    else:
                        # New message for same archive(s) as held over
                        if self._heldover[0] == 1:
                            # Just holding a single message so insert it into a
                            # temporary mbox file
                            mbox_path = tempfile.mktemp('.mbox')
                            mbox_fp = open(mbox_path, 'w')
                            mbox_fp.write(self._heldover[2].as_string(True) + '\n')
                            self._heldover[2] = mbox_path
                            self._heldover[0] = 2
                        # Add new message to the held over temporary mbox file
                        mbox_fp.write(msg.as_string(True) + '\n')
        finally:
            # Always close the temporary mbox file on the way out
            if mbox_fp:
                mbox_fp.close()
        return result
